#!/usr/bin/python3

caps = [ord(c) for c in "ABCDEFGHIKLMNOPRSTUVWXYZ"]
xchars = [ord(c) for c in "acemnorsuvwxz"]

def convert_def(fn):
	src = open(fn)

	result = ""
	glyphs = {}
	metrics = {}
	kerning = []
	codes = {}
	cap_height = []
	x_height = []
	for line in src.readlines():
		line = line.strip()
		if not line or line[0]=='#':
			continue

		parts = line.split()
		values = tuple(map(int, parts[1:]))

		if parts[0]=="font":
			tw, th, fh, fa, fd = map(int, parts[1:])
			result += "native_size {};\n".format(fh)
			result += "ascent {:.3f};\n".format(float(fa)/fh)
			result += "descent {:.3f};\n".format(float(fd)/fh)
		elif parts[0]=="code":
			c, n = tuple(map(int, parts[1:]))
			codes[c] = n
		elif parts[0]=="metrics":
			metrics[values[0]] = values[1:]
		elif parts[0]=="glyph":
			glyphs[values[0]] = values[1:]
		elif parts[0]=="kern":
			kerning.append(values)

	for c, n in codes.items():
		h = metrics[n][1]
		if c in caps:
			cap_height.append(h)
		elif c in xchars:
			x_height.append(h)

	if cap_height:
		cap_height.sort()
		result += "cap_height {:.3f};\n".format(float(cap_height[len(cap_height)*2//3])/fh)
	if x_height:
		x_height.sort()
		result += "x_height {:.3f};\n".format(float(x_height[len(x_height)*2//3])/fh)

	for c, n in codes.items():
		_, _, ox, oy, a = metrics[n]
		x, y, w, h, b = glyphs[n]
		result += "glyph {}\n{{\n".format(c)
		result += "\ttexcoords {} {} {} {};\n".format(float(x)/tw, float(th-y-h)/th, float(x+w)/tw, float(th-y)/th)
		result += "\tsize {:.3f} {:.3f};\n".format(float(w)/fh, float(h)/fh)
		result += "\toffset {:.3f} {:.3f};\n".format(float(ox-b)/fh, float(oy-b)/fh)
		result += "\tadvance {:.3f};\n".format(float(a)/fh)
		result += "};\n"

	for k in kerning:
		l, r, d = k
		result += "kerning {} {} {:.3f};\n".format(l, r, float(d)/fh)

	return result

def make_font(fn, size, ch_ranges, autohinter, margin, padding, distfield):
	import maketex
	import os

	cmd = "ttf2png \"{}\" -o makefont-tmp.png -d makefont-tmp.def -t -p -s {} -m {} -n {} -g".format(fn, size, margin, padding)
	for r in ch_ranges:
		cmd += " -r {},{}".format(*r)
	if autohinter:
		cmd += " -a"
	if distfield:
		cmd += " -f {}".format(distfield)
	if os.system(cmd):
		raise Exception("Could not execute ttf2png")

	result = "texture\n{\n"
	result += maketex.make_tex("makefont-tmp.png", wrap="CLAMP_TO_EDGE")
	result += "};\n"
	result += convert_def("makefont-tmp.def")

	os.unlink("makefont-tmp.png")
	os.unlink("makefont-tmp.def")

	return result

def parse_range(s):
	return tuple(map(int, s.split(',', 1)))

if __name__=="__main__":
	import sys
	import argparse

	parser = argparse.ArgumentParser()
	parser.add_argument("-s", "--size", default=10, type=int, metavar="NUM", help="Font size in pixels")
	parser.add_argument("-a", "--autohinter", action="store_const", const=True, default=False, help="Force autohinter")
	parser.add_argument("-m", "--margin", default=0, type=int, metavar="NUM", help="Margin around image edge in pixels")
	parser.add_argument("-n", "--padding", default=1, type=int, metavar="NUM", help="Padding between glyphs in pixels")
	parser.add_argument("-r", "--range", action="append", type=parse_range, metavar="START,END", help="Range of code points to include")
	parser.add_argument("-f", "--distfield", default=0, type=int, metavar="NUM", help="Produce a distance field texture")
	parser.add_argument("ttf", metavar="font.ttf", help="TrueType file to read")
	parser.add_argument("outfile", help="MspGL font file to write")

	args = parser.parse_args()

	out = open(args.outfile, "w")
	out.write(make_font(args.ttf, args.size, args.range, args.autohinter, args.margin, args.padding, args.distfield))
